#' BHPathway Function
#'
#' BHPathway is used to conduct pathway analysis using a combination of Iteratively Weighted Least Squares and Expectation-Maximization algorithms. It outputs p-values at three levels, namely, pathway, genes and SNPs. This function can optionally calculate effective number of parameters for each group of pathways, group of genes, and  group of SNPs, and conduct inference based on the effective number of parameters.
#' @references Zhang, L (2018) A Bayesian Hierarchical Framework for Pathway Analysis in Genome-Wide Association Studies. Ph.D. Dissertation. The University of Texas at Dallas.
#' @param type_of_genotype can be either "allelic" such as “A G” , or "numeric", e.g., 0, 1, or 2.
#' @param geno SNP genotype matrix with one row per individual and columns listing SNPs in order. If type_of_genotype="allelic", geno has number of columns as 2*number of SNPs. If type_of_genotype="numeric", number of columns in geno is same as the number of SNPs.
#' @param response takes values of 0 or 1.
#' @param pw_gene_snp_info takes three columns, column 1 with pathway ids, column 2 with gene ids, column 3 with SNP ids. The number of rows needs to be same as the total number of SNPs.
#' @param max.iter maximum number of iterations allowed until convergence.
#' @param FDR.target user specified value of FDR boundary for conducting  inference, default value is 0.1.
#' @param hierarchical.inference whether inference should be carried out hierarchically at pathway, gene, and SNP levels. Default value is "FALSE" in which case non-hierarchical inference is carried out at pathway level only.
#' @param delta.star used for calculating hierarchical FDR boundary, default value is 1. Needed only when hierarchical.inference=TRUE.
#' @param use.eff this is relevant only when hierarchical.inference="TRUE". If use.eff=="TRUE", cutoffs used in the hierarchical inference in BH (Benjamini-Hochberg) procedure will be based on the effective number of parameters. If use.eff=="FALSE", hierarchical inference uses usual BH thresholds. Default value is "FALSE".
#' @return "results" is returned as a list containing following items:
#' \item{pw.pvalue}{vector of p-values for all pathways under study.}
#' \item{gene.pvalue}{vector of p-values for all genes under study.}
#' \item{snp.pvalue}{vector of p-values for all SNPs under study.}
#' \item{pw.eff.num}{ effective number of parameters for the group of pathways.}
#' \item{gene.eff.num}{ effective number(s) of parameters for the groups of genes in each pathway.}
#' \item{snp.eff.num}{ effective number(s) of parameters for the groups of SNPs in each gene.}
#' \item{pw.reject.result}{ rejection results at pathway level.}
#' \item{gene.reject.result}{ rejection results at gene level. This is relevant only when hierarchical.inference="TRUE". "Rejected" means the hypothesis is tested and rejected. "Not tested" means this gene is under a pathway which is not significant, and thus, this gene was not tested.}
#' \item{snp.reject.result}{ rejection results at SNP level. This is relevant only when hierarchical.inference="TRUE". "Not tested" means this SNP is under a gene or a pathway which is not significant, and thus, this SNP was not tested.}
#' \item{chosen.qvalue}{ the q-value such that  FDR boundary is equal to the user specified value in variable "FDR.target".}
#' @export
#' @examples
#' ## Load example data sets
#' data(pw_gene_snp_info_eg)  ## This consists of information of pathway, gene, SNP mapping information.
#' data(response_eg)  ## This consists of affection status.
#' data(geno_eg) ## This consists of SNP genotype matrix.
#'
#' ## Non-Hierarchical inference.
#' pw.level.without.eff<-BHPathway(type_of_genotype="numeric",geno=geno_eg,response=response_eg,
#'                                 pw_gene_snp_info=pw_gene_snp_info_eg,max.iter=200,
#'                                 FDR.target=0.25,hierarchical.inference="FALSE",delta.star=1,use.eff="FALSE")
#'
#' ## Hierarchical inference without effective number of parameters.
#' all.level.without.eff<-BHPathway(type_of_genotype="numeric",geno=geno_eg,response=response_eg,
#'                                  pw_gene_snp_info=pw_gene_snp_info_eg,max.iter=200,
#'                                  FDR.target=0.25,hierarchical.inference="TRUE",delta.star=1,use.eff="FALSE")
#'
#' ## Hierarchical inference with effective number of parameters.
#' all.level.with.eff<-BHPathway(type_of_genotype="numeric",geno=geno_eg,response=response_eg,
#'                               pw_gene_snp_info=pw_gene_snp_info_eg,max.iter=200,
#'                               FDR.target=0.25,hierarchical.inference="TRUE",delta.star=1,use.eff="TRUE")


BHPathway<-function(type_of_genotype="numeric",geno,response,pw_gene_snp_info,max.iter=200,
                    FDR.target=0.1,hierarchical.inference="FALSE",delta.star=1,use.eff="FALSE")
{

  eff.pw<-function(prior.variance){
    B.for.pw=chol2inv(chol(chol2inv(chol(C4))+t(A3)%*%t(A2)%*%t(A1)%*%chol2inv(chol(C1+A1%*%C2%*%t(A1)+A1%*%A2%*%C3%*%t(A2)%*%t(A1)))%*%A1%*%A2%*%A3))
    left=length(prior.variance)-sum((1/prior.variance)*diag(B.for.pw))
    return(left)
  }

  eff.gene<-function(prior.variance){
    A1.mu=A1[,-c(s.start:s.end)]#==head( A1[,7:21])
    A2.mu=A2[-c(s.start:s.end),-c(gstart:gend)] #==  A2[,1:2]
    C3.mu=C3[-c(gstart:gend),-c(gstart:gend)]

    B.for.gene=chol2inv(chol(chol2inv(chol(C3[gstart:gend,gstart:gend]))+t(A1[,s.start:s.end]%*%A2[s.start:s.end,gstart:gend])%*%chol2inv(chol(C1+A1%*%C2%*%t(A1)+A1.mu%*%A2.mu%*%C3.mu%*%t(A2.mu)%*%t(A1.mu)))%*%A1[,s.start:s.end]%*%A2[s.start:s.end,gstart:gend]))
    left = length(prior.variance)-sum((1/prior.variance)*diag(B.for.gene))
    return (left)
  }

  eff.snp<-function(prior.variance){
    A1.snp=as.matrix(A1[,snp.start:snp.end]);dim(A1.snp)
    C2.snp=as.matrix(C2[snp.start:snp.end,snp.start:snp.end]);dim(C2.snp)
    B.for.snp=chol2inv(chol(chol2inv(chol(C2.snp))+t(A1.snp)%*%(C1+A1%*%C2%*%t(A1)-A1.snp%*%C2.snp%*%t(A1.snp))%*%A1.snp))
    left=length(prior.variance)-sum((1/prior.variance)*diag(B.for.snp))
    return (left)
  }

  rej.eff <- function(pvalue,qvalue,eff_num){
    ordered=order(pvalue,decreasing=TRUE) # give index of ordered values of object
    sorted=sort(pvalue,decreasing=TRUE)  #  order only the values of object
    reject=rep("Not Rejected",length(sorted))
    for (i in 1:length(sorted)){
      proportion<-(length(sorted)-i+1)/eff_num
      if(sorted[i]<qvalue*min(proportion,1)){
        reject[ordered[i:length(sorted)]]="Rejected";break
      }
    }
    return(reject)
  }

  rej <- function(pvalue,qvalue){
    ordered=order(pvalue,decreasing=TRUE) # give index of ordered values of object
    sorted=sort(pvalue,decreasing=TRUE)  #  order only the values of object
    reject=rep("Not Rejected",length(sorted))
    for (i in 1:length(sorted)){
      if(sorted[i]<qvalue*(length(sorted)-i+1)/length(sorted) ){
        reject[ordered[i:length(sorted)]]="Rejected";break
      }
    }
    return(reject)
  }

  #########if readin data is in terms of pair of allele
  if (type_of_genotype=="allelic"){

    geno<-as.matrix(geno)
    count.minor <- function(input){
      (level <- table(input))
      (name <- names(level))
      if(length(level)!=1){
        if(level[1]==level[2]){
          major <- name[1]
          minor <- name[2]
        }

        if(level[1]!=level[2]){
          (major <- name[which(level==max(level))])
          (minor <- name[which(level==min(level))])
        }

        (converted<-replace(input,input==minor,1))
        (converted<-replace(converted,input==major,0))
        (converted<-apply(converted,2,as.numeric))
      }

      if(length(level)==1){
        major=name[1];minor=name[1]
        (converted<-replace(input,input==major,0))  #because for most of the cases, if only "A,A" occurs , "A" is most likely to be major allele. So genotype should be 0
        (converted<-apply(converted,2,as.numeric))
      }

      (result=apply(converted,1,sum))
      return(t(t(result)))
    }
    genotype.to.change <- matrix(NA,nrow=nrow(geno),ncol=ncol(geno)/2)
    for (i in 1:ncol(genotype.to.change)){
      input<-geno[,(2*i-1):(2*i)]
      genotype.to.change[,i]<-count.minor(input)
    }
    #genotype.to.change[,tochange]<-(2-genotype.to.change[,tochange])
    genotype<-as.matrix(genotype.to.change)
  } else if (type_of_genotype=="numeric"){
    genotype<-as.matrix(geno)
  } else{
    print("Genotype format not able to handle.");break
  }
  ###################################################################
  y<-response
  pw_gene_snp_info<-as.matrix(pw_gene_snp_info)
  ####################################################################
  #set values of tuning parameters
  v_lambda=1; s_lambda=2.5;
  w.tau=0.04; a=6.25;b=1;
  u= 1;v=1
  #################################################
  K=length(unique(pw_gene_snp_info[,1]))
  NGENE<-rep(NA,K);NSNP<-NA
  for (k in 1:K){
    index<-which(pw_gene_snp_info[,1]==unique(pw_gene_snp_info[,1])[k])
    NGENE[k]=length(unique(pw_gene_snp_info[index,2]))
    for (g in 1:NGENE[k]){
      NSNP.current<-sum(pw_gene_snp_info[index,2]==unique(pw_gene_snp_info[index,2])[g])
      NSNP<-c(NSNP,NSNP.current)
    }
  }  ; NSNP<-NSNP[-1]
  (G=sum(NGENE)  )    #   number of genes in each pathway
  (J=sum(NSNP)); NJ=J+1 #  number of snps in each gene, J is total # of snps, NJ includes intercept
  NGENE.extend=c(1,NGENE);NSNP.extend=c(1,NSNP)
  ##############################################################################
  #this record the membership of snps to gene,gene to pathway
  D22=matrix(0,nrow=NJ,ncol=G);  D22[1,]=0 ; D22[2:(1+NSNP[1]),1]=-1
  for ( i in 1:(G-1) ){   D22[(2+sum(NSNP[1:i])):(1+sum(NSNP[1:(i+1)])),i+1]=-1  }
  D33=matrix(0,nrow=G,ncol=K);
  for ( k in 1:K){   D33[(sum(NGENE.extend[1:k])):(sum(NGENE.extend[2:(k+1)])),k]=-1   }
  #####################################################
  ###  Setting up design matrix
  conv <- FALSE;nobs <- NROW(y)
  D.nrow<-(nobs+NJ+G+K);  D.ncol<-(NJ+G+K)
  D<-matrix(0,nrow=D.nrow,ncol=D.ncol)
  D[1:nobs,1]=1;   D[1:nobs,2:NJ]=as.matrix(genotype);    D[1:nobs,(NJ+1):(NJ+G)]=0;   D[1:nobs,(NJ+G+1):(NJ+G+K)]=0
  D[(nobs+1):(nobs+NJ),1:NJ]=diag(NJ);    D[(nobs+1):(nobs+NJ),(NJ+1):(NJ+G)]= D22;   D[(nobs+1):(nobs+NJ),(NJ+G+1):(NJ+G+K)]=0;
  D[(nobs+NJ+1):(nobs+NJ+G),1:NJ]=0;      D[(nobs+NJ+1):(nobs+NJ+G),(NJ+1):(NJ+G)]= diag(G);     D[(nobs+NJ+1):(nobs+NJ+G),(NJ+G+1):(NJ+G+K)]=D33;
  D[(nobs+NJ+G+1):(nobs+NJ+G+K),1:NJ]= 0; D[(nobs+NJ+G+1):(nobs+NJ+G+K),(NJ+1):(NJ+G)]=0;        D[(nobs+NJ+G+1):(nobs+NJ+G+K),(NJ+G+1):(NJ+G+K)]=diag(K);
  gc()
  ##############################################################################
  lambda=10 #just an initial value
  start=rep(0.0001,NJ+G+K)
  etastart=NULL;   mustart=NULL;   control=glm.control(); weights=rep(1,nobs)

  family=binomial(link="logit")
  if (is.null(weights))
    weights <- rep.int(1, nobs)
  if (is.null(offset))   offset <- rep.int(0, nobs)
  variance <- family$variance
  dev.resids <- family$dev.resids
  linkinv <- family$linkinv
  mu.eta <- family$mu.eta
  if (!is.function(variance) || !is.function(linkinv)) stop("'family' argument seems not to be a valid family object")

  eta = D[1:nobs,] %*% start
  mu =linkinv(eta)
  devold <- sum(dev.resids(y, mu, weights))
  boundary <- conv <- FALSE
  ######### assigning initial values on prior means and variances
  gene.mu=rep(1,G) ; gene.sigmasquare=rep(2,G)
  # initial values on prior means and variances for snps
  beta.prior.mean=intercept.mean=0  ;   beta.prior.variance=intercept.variance=2;
  for (i in 1:G) { beta.prior.mean=append(beta.prior.mean,rep(gene.mu[i],NSNP[i]));
  beta.prior.variance=append(beta.prior.variance,rep(gene.sigmasquare[i],NSNP[i]))
  }

  # initial values on prior means and variances for genes
  pw.theta=rep(1,K); pw.ssquare= rep(2,K);  pw.tausquare=rep(2,K)
  mu.prior.mean=rep(pw.theta[1], NGENE[1])  #number of genes in 1st pathway
  mu.prior.variance=rep(pw.tausquare[1], NGENE[1])

  sigma.prior.scale=intercept.hyper=10^2 # this dose not need to be updated
  for (i in 1:K) {
    sigma.prior.scale=append(sigma.prior.scale, rep(pw.ssquare[i], NGENE[i] )) # this records all the prior variance (s-square) for variance parameters of sigma square for intercept and snps
    if(i>1) {
      mu.prior.mean=append(mu.prior.mean, rep(pw.theta[i], NGENE[i])) # this completes the vector of prior mean of genes
      mu.prior.variance= append(mu.prior.variance,rep(pw.tausquare[i], NGENE[i])) # this completes the vector of prior variance of genes
    }
  }

  theta.prior.mean<-rep(0,K); theta.prior.variance<-rep(lambda^2,K)   # prior mean  and variance of pw won't change

  max.iter<-max.iter
  for (iteration in 1:max.iter) {
    weights <- rep.int(1, nobs)
    good <- weights > 0
    varmu <- variance(mu)[good]
    if (any(is.na(varmu)))
      stop("NAs in V(mu)")
    if (any(varmu == 0))
      stop("0s in V(mu)")
    mu.eta.val <- mu.eta(eta)
    if (any(is.na(mu.eta.val[good])))
      stop("NAs in d(mu)/d(eta)")
    good <- (weights > 0) & (mu.eta.val != 0)
    if (all(!good)) {
      conv <- FALSE
      warning("no observations informative at iteration ",
              iter)
      break
    }
    z = eta  + (y - mu)/mu.eta.val                   ##pseudo data z1....zn
    w = sqrt((weights * mu.eta.val^2)/variance(mu))  ##square root of pseudo weight  w1.....wn

    # This is where we augment the data with the prior information
    z.extend=rep(0,NJ+G+K) ; length(z.extend) ##In addition to pesudo-data z, we had response 0's for our extened model
    z.star <- c(z,z.extend) ; length(z.star) ;
    x.star <- D         ##extended design matrix
    w.star <- c(w, 1/sqrt(beta.prior.variance),1/sqrt(mu.prior.variance),1/sqrt(theta.prior.variance))

    x.prior=D[(nobs+1):(nobs+NJ+G+K),]
    good.star=c(good,rep(TRUE,NCOL(x.prior)))

    fit=qr(x.star[good.star,,drop=FALSE]*w.star,tol=min(10^(-7),control$epsilon/1000))
    fit$coefficients=qr.coef(fit,z.star*w.star)  # regression for coefficients
    beta=fit$coefficients[1:NJ];                 #estimated intercept and coefficients for snps
    gene.mu=fit$coefficients[(NJ+1):(NJ+G)];     #estimated coefficient for gene
    pw.theta=fit$coefficients[(NJ+G+1):(NJ+G+K)] #estimated coefficient for pw

    if(any(fit$coefficients>(10^5))){
      cat("Infinity parameter estimation reached")
      break
    }

    fit$qr=as.matrix(fit$qr)
    COV=chol2inv(fit$qr[1:ncol(x.star),1:ncol(x.star),drop=FALSE])
    V.para=diag(chol2inv(fit$qr[1:ncol(x.star),1:ncol(x.star),drop=FALSE])); #joint posterior covariance matrix for all parameters

    V.for.beta=V.para[1:NJ] ; V.for.mu=V.para[(NJ+1):(NJ+G)] ; V.for.theta=V.para[(NJ+G+1):(NJ+G+K)]

    ######## now updated the unknown hyper-parameters
    intercept.variance=(v*intercept.hyper+beta[1]^2+V.for.beta[1])/(2+v+1)  #update step 1
    #updating part 1 : updating sigma square
    for (g in 1:G){
      snp.start=sum(NSNP.extend[1:g])
      snp.end=sum(NSNP.extend[2:(g+1)])
      gene.sigmasquare[g]=(
        v*sigma.prior.scale[g+1]+sum((beta[(1+snp.start):(1+snp.end)]-gene.mu[g])^2+V.for.beta[(1+snp.start):(1+snp.end)])
        +NSNP[g]*V.for.mu[g]-2*sum(COV[(NJ+g),(snp.start+1):(snp.end+1)])
      )/(v+2+NSNP[g])
    }

    for (k in 1:K) {
      gstart=sum(NGENE.extend[1:k])
      gend=sum(NGENE.extend[2:(k+1)])

      #updating part 2  updating s square #corrected with no 1/2 in the denominator
      pw.ssquare[k]=(v*NGENE[k]+2*a-2)/(2*b+ v*sum(1/gene.sigmasquare[gstart:gend]))

      #updating part 3  updating tau square
      pw.tausquare[k]=(
        u*(w.tau^2)+sum((gene.mu[gstart:gend]-pw.theta[k])^2+V.for.mu[gstart:gend])
        +NGENE[k]*V.for.theta[k]-2*sum(COV[(NJ+G+k),(NJ+gstart):(NJ+gend)])
      )/(NGENE[k]+u+2)
    }

    ###updating lambda
    lambda<-sqrt((sum(pw.theta^2+V.for.theta)+v_lambda*(s_lambda^2))/(K+2+v_lambda))


    ##with unknown parameter updated, prior means and variances are updated through below
    beta.prior.mean=intercept.mean=0  ; beta.prior.variance=intercept.variance
    for (i in 1:G) { beta.prior.mean=append(beta.prior.mean,rep(gene.mu[i],NSNP[i]));
    beta.prior.variance=append(beta.prior.variance,rep(gene.sigmasquare[i],NSNP[i]))
    }

    sigma.prior.scale=intercept.hyper
    mu.prior.mean=rep(pw.theta[1], NGENE[1])  #***number of genes in 1st pathway
    mu.prior.variance=rep(pw.tausquare[1], NGENE[1])
    for (i in 1:K) { #continue with rest of the pathways
      sigma.prior.scale=append(sigma.prior.scale, rep(pw.ssquare[i], NGENE[i] ) )
      if(i>1) {
        mu.prior.mean=append(mu.prior.mean, rep(pw.theta[i], NGENE[i]))
        mu.prior.variance= append(mu.prior.variance,rep(pw.tausquare[i], NGENE[i]))
      }
    }

    theta.prior.variance=rep(lambda^2,K)
    #print(lambda)
    ############# After one iteration of regression, estimated coefficients then become the starting value for updating missing parameters
    start=fit$coefficients
    eta=drop(D[1:nobs,]%*%start)
    mu=linkinv(eta); mu
    dev=sum(dev.resids(y,mu,weights))

    # if converge criteria is met, regression part will break, otherwise continue
    if (iteration>1 & abs(dev-devold)/(0.1+abs(dev))<control$epsilon){
      conv=TRUE;coef=start;
      #cat("Algorithm converge at iteration:",iteration,"\n")
      break
    }else {  devold=dev;iteration=iteration+1}
    gc()
  }
  #coefficients estimation done
  #########################################################################################
  stat.beta=beta[2:NJ]/sqrt(V.for.beta[2:NJ])
  qr.snp.pvalue=2*pnorm(abs(stat.beta),lower.tail=FALSE)

  stat.mu=gene.mu/sqrt(V.for.mu)
  qr.gene.pvalue=2*pnorm(abs(stat.mu),lower.tail=FALSE)

  stat.theta=pw.theta/sqrt(V.for.theta)
  qr.pw.pvalue=2*pnorm(abs(stat.theta),lower.tail = FALSE)
  #calculation for effective number for all groups start here
  eff.for.pw=rep(NA,K); eff.for.gene=rep(NA,K) ; eff.for.snp=rep(NA,G)

  A1=D[1:length(y),2:NJ]; A2=-D22[2:NJ,]; A3=-D33
  C1=diag(1/(w^2))  # use w^2 for qr decomposition, use w for direct calculation
  C2=diag(beta.prior.variance[2:NJ]); C3=diag(mu.prior.variance);
  C4=diag(lambda^2,K)

  eff.for.pw=eff.pw(theta.prior.variance)
  for (k in 1:K){
    gstart=sum(NGENE.extend[1:k])
    gend=sum(NGENE.extend[2:(k+1)])
    s.start=sum(NSNP.extend[1:gstart])
    s.end=sum(NSNP.extend[2:(gend+1)])
    eff.for.gene[k]=eff.gene(mu.prior.variance[gstart:gend])
    for (g in sum(NGENE.extend[1:k]):sum(NGENE.extend[2:(k+1)])) {
      snp.start=sum(NSNP.extend[1:g])
      snp.end=sum(NSNP.extend[2:(g+1)])
      eff.for.snp[g]=eff.snp(beta.prior.variance[(1+snp.start):(1+snp.end)])
    }
  }
  #calculation for effective number done

  pw_pvalue<-qr.pw.pvalue;gene_pvalue<-qr.gene.pvalue;snp_pvalue<-qr.snp.pvalue;
  pw_eff<-eff.for.pw;gene_eff<-eff.for.gene;snp_eff<-eff.for.snp;

  ###now conduct inference
  if (hierarchical.inference=="FALSE"&&use.eff=="TRUE"){
    qvalue<-FDR.target
    pw.reject.eff<-rej.eff(pw_pvalue,qvalue,pw_eff)

    pw.reject.result<-pw.reject.eff
    gene.reject.result<-"Not offered when hierarchical.inference is FALSE"
    snp.reject.result<-"Not offered when hierarchical.inference is FALSE"


  }else if (hierarchical.inference=="FALSE"&&use.eff=="FALSE"){
    qvalue<-FDR.target
    pw.reject<-rej(pw_pvalue,qvalue)

    pw.reject.result<-pw.reject
    gene.reject.result<-"Not offered when hierarchical.inference is FALSE"
    snp.reject.result<-"Not offered when hierarchical.inference is FALSE"

  }else if (hierarchical.inference=="TRUE"&&use.eff=="FALSE"){
    hFDR.target<-FDR.target
    qval<-seq(0,0.5,0.01);fdr_boundry<-rep(NA,length(qval))
    for (i in 1:length(qval)){
      qvalue=qval[i];
     # print(qvalue)
      ############################################################################
      pw.reject<-rep('Not Rejected',K);gene.reject<-rep('Not Rejected',G);snp.reject<-rep('Not Rejected',J)
      pw.reject<-rej(pw_pvalue,qvalue)

      for (k in 1:K){
        if(pw.reject[k]=="Rejected"){
          gstart<-sum(NGENE.extend[1:k])
          gend<-sum(NGENE.extend[2:(k+1)])
          s.start<-sum(NSNP.extend[1:gstart])
          s.end<-sum(NSNP.extend[2:(gend+1)])
          gene.reject[gstart:gend]<-rej(gene_pvalue[gstart:gend],qvalue)

          for (g in gstart:gend) {
            if(gene.reject[g]=="Rejected"){
              snp.start<-sum(NSNP.extend[1:g])
              snp.end<-sum(NSNP.extend[2:(g+1)])
              snp.reject[snp.start:snp.end]<-rej(as.numeric(unlist(snp_pvalue))[snp.start:snp.end],qvalue)
            }
          }
        }
      }
      if(sum(is.na(snp.reject))<length(snp.reject)){
        total.sig=sum(pw.reject=="Rejected",gene.reject=="Rejected",
                      snp.reject=="Rejected")
      }else{
        total.sig=sum(pw.reject=="Rejected",gene.reject=="Rejected")
      }

      total.family=1+sum(pw.reject=="Rejected",gene.reject=="Rejected")

      multiplier=(total.sig+total.family)/(1+total.sig)
      #print(multiplier)
      (fdr_boundry[i]<-qvalue*delta.star*multiplier)
    }
    dex=max(which(abs(fdr_boundry-hFDR.target)==min(abs(fdr_boundry-hFDR.target)))) # find which qvalue corresponds to FDR closest to 0.05
    qvalue<-qval[dex]

    pw.reject<-rep("Not Rejected",K);gene.reject<-rep("Not tested",G);snp.reject<-rep("Not tested",J)
    pw.reject<-rej(pw_pvalue,qvalue)

    for (k in 1:K){
      if(pw.reject[k]=="Rejected"){
        gstart<-sum(NGENE.extend[1:k])
        gend<-sum(NGENE.extend[2:(k+1)])
        s.start<-sum(NSNP.extend[1:gstart])
        s.end<-sum(NSNP.extend[2:(gend+1)])
        gene.reject[gstart:gend]<-rej(gene_pvalue[gstart:gend],qvalue)
       # print(gene.reject)
        for (g in gstart:gend) {
          if(gene.reject[g]=="Rejected"){
            snp.start<-sum(NSNP.extend[1:g])
            snp.end<-sum(NSNP.extend[2:(g+1)])
            snp.reject[snp.start:snp.end]<-rej(as.numeric(unlist(snp_pvalue))[snp.start:snp.end],qvalue)
          }
        }
      }
    }
    if(sum(is.na(snp.reject))<length(snp.reject)){
      total.sig=sum(pw.reject=="Rejected",gene.reject=="Rejected",
                    snp.reject=="Rejected")
    }else{
      total.sig=sum(pw.reject=="Rejected",gene.reject=="Rejected")
    }
    total.family=1+sum(pw.reject=="Rejected",gene.reject=="Rejected")
    multiplier<-(total.sig+total.family)/(1+total.sig)
    empirical_fdr_boundry<-qvalue*delta.star*multiplier
    pw.reject.result<-pw.reject
    gene.reject.result<-gene.reject
    snp.reject.result<-snp.reject

  } else if (hierarchical.inference=="TRUE"&&use.eff=="TRUE"){
    hFDR.target<-FDR.target
    qval<-seq(0,0.5,0.01);fdr_boundry<-rep(NA,length(qval))
    for (i in 1:length(qval)){
      qvalue=qval[i];
     # print(qvalue)
      ############################################################################
      pw.reject.eff<-rep("Not Rejected",K);gene.reject.eff<-rep("Not tested",G);snp.reject.eff<-rep("Not tested",J)
      pw.reject.eff<-rej.eff(pw_pvalue,qvalue,pw_eff)

      for (k in 1:K){
        if(pw.reject.eff[k]=="Rejected"){
          gstart<-sum(NGENE.extend[1:k])
          gend<-sum(NGENE.extend[2:(k+1)])
          s.start<-sum(NSNP.extend[1:gstart])
          s.end<-sum(NSNP.extend[2:(gend+1)])
          gene.reject.eff[gstart:gend]<-rej.eff(gene_pvalue[gstart:gend],qvalue,gene_eff[k])
         # print(gene.reject.eff)
          for (g in gstart:gend) {
            if(gene.reject.eff[g]=="Rejected"){
              snp.start<-sum(NSNP.extend[1:g])
              snp.end<-sum(NSNP.extend[2:(g+1)])
              snp.reject.eff[snp.start:snp.end]<-rej.eff(as.numeric(unlist(snp_pvalue))[snp.start:snp.end],qvalue,snp_eff[g])
            }
          }
        }
      }
      if(sum(is.na(snp.reject.eff))<length(snp.reject.eff)){
        total.sig=sum(pw.reject.eff=="Rejected",gene.reject.eff=="Rejected",
                      snp.reject.eff=="Rejected")
      }else{
        total.sig=sum(pw.reject.eff=="Rejected",gene.reject.eff=="Rejected")
      }

      total.family=1+sum(pw.reject.eff=="Rejected",gene.reject.eff=="Rejected")

      multiplier=(total.sig+total.family)/(1+total.sig)
     # print(multiplier)
      (fdr_boundry[i]<-qvalue*delta.star*multiplier)
    }
    dex=max(which(abs(fdr_boundry-hFDR.target)==min(abs(fdr_boundry-hFDR.target)))) # find which qvalue corresponds to FDR closest to 0.05
    qvalue<-qval[dex]
    ############################################################################
    pw.reject.eff<-rep("Not Rejected",K);gene.reject.eff<-rep("Not tested",G);snp.reject.eff<-rep("Not tested",J)
    pw.reject.eff<-rej.eff(pw_pvalue,qvalue,pw_eff)

    for (k in 1:K){
      if(pw.reject.eff[k]=="Rejected"){
        gstart<-sum(NGENE.extend[1:k])
        gend<-sum(NGENE.extend[2:(k+1)])
        s.start<-sum(NSNP.extend[1:gstart])
        s.end<-sum(NSNP.extend[2:(gend+1)])
        gene.reject.eff[gstart:gend]<-rej.eff(gene_pvalue[gstart:gend],qvalue,gene_eff[k])
       # print(gene.reject.eff)
        for (g in gstart:gend) {
          if(gene.reject.eff[g]=="Rejected"){
            snp.start<-sum(NSNP.extend[1:g])
            snp.end<-sum(NSNP.extend[2:(g+1)])
            snp.reject.eff[snp.start:snp.end]<-rej.eff(as.numeric(unlist(snp_pvalue))[snp.start:snp.end],qvalue,snp_eff[g])
          }
        }
      }
    }
    if(sum(is.na(snp.reject.eff))<length(snp.reject.eff)){
      total.sig=sum(pw.reject.eff=="Rejected",gene.reject.eff=="Rejected",
                    snp.reject.eff=="Rejected")
    }else{
      total.sig=sum(pw.reject.eff=="Rejected",gene.reject.eff=="Rejected")
    }
    total.family=1+sum(pw.reject.eff=="Rejected",gene.reject.eff=="Rejected")
    multiplier<-(total.sig+total.family)/(1+total.sig)
    empirical_fdr_boundry<-qvalue*delta.star*multiplier
    pw.reject.result<-pw.reject.eff
    gene.reject.result<-gene.reject.eff
    snp.reject.result<-snp.reject.eff
  }


  #list outputs
  pw.pvalue=qr.pw.pvalue
  results<-list(pw.pvalue=qr.pw.pvalue,gene.pvalue=qr.gene.pvalue,snp.pvalue=qr.snp.pvalue,
                pw.eff.num=eff.for.pw,gene.eff.num=eff.for.gene,snp.eff.num=eff.for.snp,
                pw.reject.result=pw.reject.result,gene.reject.result=gene.reject.result,snp.reject.result=snp.reject.result,
                chosen.qvalue=qvalue)
  return(c(pw.pvalue,results))

}
